#!/bin/python3.6

# external
import os
import shutil
from subprocess import Popen
import sys
import networkx as nx
import random
from typing import Dict, List, Tuple, Callable
from pprint import pprint
from time import sleep, perf_counter as now
import argparse
import signal
from collections import Counter
from operator import itemgetter
from functools import reduce
from heapq import heappop, heappush
from dataclasses import dataclass, field

# internal
from blip import (
    run_blip,
    BayesianNetwork,
    TWBayesianNetwork,
    monitor_blip,
    parse_res,
    start_blip_proc,
    check_blip_proc,
    stop_blip_proc,
    write_res,
    activate_checkpoints,
    run_hcet,
    start_hcet_proc,
    check_hcet_proc,
)
from utils import (
    TreeDecomposition,
    pick,
    pairs,
    filter_read_bn,
    BNData,
    NoSolutionError,
    UnsatisfiableInstanceError,
    get_domain_sizes,
    log_bag_metrics,
    compute_complexity_width,
    weight_from_domain_size,
    compute_complexity,
    compute_complexities,
    shuffled,
)
from berg_encoding import solve_bn, PSET_ACYC

# optional
try:
    import wandb
except ImportError:
    wandb = None


# check if drawing/plotting is available
CLUSTER = os.environ.get("WANDB_PLATFORM", "cluster") == "cluster"
if not CLUSTER:
    try:
        from networkx.drawing.nx_agraph import graphviz_layout
        import matplotlib.pyplot as plt
    except ImportError:
        CLUSTER = False
        graphviz_layout = None
        plt = None


# score comparison epsilon
EPSILON = 1e-7

# default parameter values
BUDGET = 10  # maximum size of local subinstance
TIMEOUT = 2  # timeout per maxsat call
MAX_PASSES = 100000  # maximum number of passes to perform
MAX_TIME = 1800  # maximum time limit
HEURISTIC = "kmax"
OFFSET = 10  # duration for which to run the heuristic
SEED = 1  # seed used for both BN-SLIM and k-MAX (if applicable)
LAZY_THRESHOLD = 0.0
START_WITH = None  # file where initial solution is stored (optional)
RELAXED_PARENTS = False
TRAV_STRAT = "random"
MIMIC = False
ANYTIME_HEURISTIC = False
MIMIC_SCORE_THRESHOLD = 1.0
DOMAIN_SIZES: Dict[int, int] = None
DATFILE = ""
USE_COMPLEXITY_WIDTH = False
USING_COMPLEXITY_WIDTH = False
COMPLEXITY_BOUND = -1
CW_TARGET_REACHED = False
CW_TRAV_STRAT = "max-rand"  # one of max, max-min, max-rand, tw-max-rand
CW_EXP_STRAT = "min-int"  # one of max, min, min-int
SAVE_AS = ""  # pattern to use while saving networks
LOGGING = False  # wandb logging
MAXSAT_SOLVER = "nuwls-c"
INTERLEAVED = False  # whether to alternate between slim and heuristic
INTERLEAVABLE_HEURISTICS = ("hc", "hcp")
INTERLEAVE_PASSES = 5
INTERLEAVE_TIMEOUT = 10  # min time to run each interleave phase of slim for (s)
AUGMENTED_JKL = "augmented.jkl"  # file to accumulate augmented psets


def find_start_bag(td: TreeDecomposition, history: Counter = None, debug=False):
    if USING_COMPLEXITY_WIDTH:  # pick bag with highest complexity
        complexities = compute_complexities(td, DOMAIN_SIZES)
        if CW_TRAV_STRAT == "max":
            bag_order = sorted(complexities, key=complexities.get, reverse=True)
        elif CW_TRAV_STRAT == "max-min":
            bag_order = sorted(
                complexities,
                key=complexities.get,
                reverse=not CW_TARGET_REACHED,
            )
        else:  # max-rand or tw-max-rand
            if CW_TARGET_REACHED:
                bag_order = shuffled(complexities.keys())
            else:
                bag_order = sorted(
                    complexities, key=complexities.get, reverse=True
                )
        mincount = min(history[bag_id] for bag_id in bag_order)
        for bag_id in bag_order:
            if history[bag_id] == mincount:
                return bag_id
        # maxbagidx, _ = max(complexities.items(), key=itemgetter(1))
        # if history[maxbagidx] == 0:
        #     return maxbagidx
        # else:  # cw target already met, return random bag
        #     print("randomly picking bag:", end="")
        #     return pick(td.bags.keys())
    elif TRAV_STRAT == "random":  # randomly pick a bag
        return pick(td.bags.keys())
    else:  # pick bag with least count and earliest in order
        if TRAV_STRAT == "post":
            trav_order = list(nx.dfs_postorder_nodes(td.decomp))
        elif TRAV_STRAT == "pre":
            trav_order = list(nx.dfs_postorder_nodes(td.decomp))
        else:
            raise ValueError(f"invalid traversal strategy {TRAV_STRAT}")
        mincount = min(history.values())
        # todo[opt]: avoid retraversing every time
        for bag_id in trav_order:
            if history[bag_id] == mincount:
                return bag_id


def find_subtree(
    td: TreeDecomposition, budget: int, history: Counter = None, debug=False
):
    """
    finds a subtree that fits within the budget

    :param td: tree decomposition in which to find the subtree
    :param budget: max number of vertices allowed in the union of selected bags
    :param history: tally of bags picked in previous iterations
    :param debug: debug mode
    :return: (selected_bag_ids, seen_vertices)
    """
    start_bag_id = find_start_bag(td, history, debug)
    selected = {start_bag_id}
    seen = set(td.bags[start_bag_id])
    if debug:
        print(f"starting bag {start_bag_id}: {td.bags[start_bag_id]}")
    queue = [(0, start_bag_id)]  # (sorting metric, bag_id)
    visited = set()
    while queue:
        _, bag_id = heappop(queue)
        visited.add(bag_id)
        bag = td.bags[bag_id]
        if len(seen.union(bag)) > budget:
            continue
        else:  # can include bag in local instance
            selected.add(bag_id)
            seen.update(bag)
            # add neighboring bags to queue
            for nbr_id in td.decomp.neighbors(bag_id):
                if nbr_id not in visited:
                    nbr_bag = td.bags[nbr_id]
                    if not USING_COMPLEXITY_WIDTH:
                        expansion_metric = (
                            0  # no sorting, so set same value for all bags
                        )
                        # todo[feature]: set to random value to randomize expansion
                    else:
                        if CW_EXP_STRAT == "max":
                            expansion_metric = -compute_complexity(
                                nbr_bag, DOMAIN_SIZES
                            )
                        elif CW_EXP_STRAT == "min":
                            expansion_metric = compute_complexity(
                                nbr_bag, DOMAIN_SIZES
                            )
                        else:
                            expansion_metric = bag.intersection(nbr_bag)
                    heappush(queue, (expansion_metric, nbr_id))
            if debug:
                print(f"added bag {bag_id}: {td.bags[bag_id]}")
    if debug:
        print(f"final seen: {seen}")
    return selected, seen


def handle_acyclicity(
    bn: BayesianNetwork, seen: set, leaf_nodes: set, debug=False
):
    dag = bn.dag
    subdag = nx.subgraph_view(
        dag, lambda x: True, lambda x, y: not ((x in seen) and (y in seen))
    )
    forced_arcs = []
    for src, dest in pairs(leaf_nodes):
        if nx.has_path(subdag, src, dest):
            forced_arcs.append((src, dest))
            if debug:
                print(f"added forced {src}->{dest}")
        else:
            # only check if prev path not found
            if nx.has_path(subdag, dest, src):
                forced_arcs.append((dest, src))
                if debug:
                    print(f"added forced {dest}->{src}")
    return forced_arcs


def prepare_subtree(
    bn: TWBayesianNetwork, bag_ids: set, seen: set, debug=False
):
    # compute leaf nodes (based on intersection of leaf bags with outside)
    boundary_nodes = set()
    forced_cliques: Dict[int, frozenset] = dict()
    for _, nbrs in bn.td.get_boundary_intersections(bag_ids).items():
        for nbr_id, intersection in nbrs.items():
            boundary_nodes.update(intersection)
            forced_cliques[nbr_id] = intersection
            if USING_COMPLEXITY_WIDTH and intersection:
                clique_cw = reduce(
                    lambda x, y: x * y, map(DOMAIN_SIZES.get, intersection)
                )
                if clique_cw >= COMPLEXITY_BOUND:
                    if debug:
                        print(
                            " skipping because marker clique exceeds cw bound"
                        )
                    SOLUTION.skipped += 1
                    return None
    if debug:
        print("clique sets:", [set(c) for c in forced_cliques.values()])

    # compute forced arc data for leaf nodes
    if debug:
        print("boundary nodes:", boundary_nodes)
    forced_arcs = handle_acyclicity(bn, seen, boundary_nodes, debug)
    if debug:
        print("forced arcs", forced_arcs)

    data, pset_acyc = get_data_for_subtree(
        bn, boundary_nodes, seen, forced_arcs
    )

    return forced_arcs, forced_cliques, data, pset_acyc


def get_data_for_subtree(
    bn: TWBayesianNetwork,
    boundary_nodes: set,
    seen: set,
    forced_arcs: List[Tuple[int, int]],
) -> Tuple[BNData, PSET_ACYC]:
    # store downstream relations in a graph
    downstream_graph = nx.DiGraph()
    downstream_graph.add_nodes_from(boundary_nodes)
    for node in boundary_nodes:
        for _, successors in nx.bfs_successors(bn.dag, node):
            downstream_graph.add_edges_from(
                (node, succ) for succ in successors if succ not in seen
            )
    # downstream.remove_nodes_from(seen - boundary_nodes)  # ignore inner red nodes
    assert seen.intersection(downstream_graph.nodes).issubset(
        boundary_nodes
    ), "downstream connectivity graph contains inner nodes"
    downstream_graph.add_edges_from(forced_arcs)

    downstream = set()
    assert RELAXED_PARENTS, "need relaxed parents always"  # todo: remove later
    if not RELAXED_PARENTS:
        downstream = set(downstream_graph.nodes()) - seen

    # construct score function data for local instance
    data: BNData = {node: dict() for node in seen}
    pset_acyc: PSET_ACYC = dict()
    for node, psets in filter_read_bn(bn.input_file, seen).items():
        if node in boundary_nodes:
            if RELAXED_PARENTS:
                downstream = set(downstream_graph.successors(node))
            for pset, score in psets.items():
                if pset.intersection(downstream):
                    continue  # reject because pset contains downstream verts
                pset_in = pset.intersection(seen)
                if len(pset_in) < len(pset):
                    # not all internal vertices, so check if bag exists
                    # todo[opt]: exclude selected while searching for bag
                    bag_id = bn.td.bag_containing(pset | {node})
                    if bag_id == -1:
                        continue  # reject because required bag doesnt already exist in td
                    rem_parents = pset - pset_in
                    req_acyc = set()
                    for parent in rem_parents:
                        if parent in downstream_graph:
                            req_acyc.update(
                                downstream_graph.predecessors(parent)
                            )
                    pset_acyc[(node, pset)] = req_acyc
                data[node][pset] = score
        else:  # internal node
            for pset, score in psets.items():
                # internal vertices are not allowed outside parents
                if pset.issubset(seen):
                    data[node][pset] = score
    return data, pset_acyc


def compute_max_score(data: BNData, bn: BayesianNetwork) -> float:
    max_score = 0
    for node in data:
        max_score += max(data[node].values()) + bn.offsets[node]
    return max_score


METRICS = (
    "start_score",
    "num_passes",
    "num_improvements",
    "skipped",
    "nosolution",
    "restarts",
    "start_width",
    "unsat",
)


class Solution(object):
    def __init__(self, value=None, logger: Callable = None):
        self.value: TWBayesianNetwork = value
        # other metrics to track
        self.data = dict.fromkeys(METRICS, 0)
        # mute logger for code completion hack
        self.logger = None
        # for code completion
        self.start_score = self.num_passes = self.num_improvements = 0
        self.skipped = self.nosolution = self.restarts = self.start_width = 0
        self.unsat = self.jumps = 0
        # set proper value for logger now
        self.logger = logger

    def update(self, new_value):
        if self.logger is not None:
            self.logger({"score": new_value.score})
            if new_value.score - 10 > self.start_score:
                self.logger({"extremely_strong": True})
            if USE_COMPLEXITY_WIDTH:
                width = compute_complexity_width(new_value.td, DOMAIN_SIZES)
                approx_width = compute_complexity_width(
                    new_value.td, DOMAIN_SIZES, approx=True
                )
                self.logger({"width": width, "approx_width": approx_width})
        self.value = new_value


def _getter_factory(metric: str):
    def getter(self: Solution):
        return self.data[metric]

    return getter


def _setter_factory(metric: str):
    def setter(self: Solution, val):
        self.data[metric] = val
        if self.logger is not None:
            self.logger({metric: val})

    return setter


for metric in METRICS:
    setattr(
        Solution,
        metric,
        property(_getter_factory(metric), _setter_factory(metric)),
    )


SOLUTION = Solution()  # placeholder global solution variable


def slimpass(
    bn: TWBayesianNetwork,
    budget: int = BUDGET,
    timeout: int = TIMEOUT,
    history: Counter = None,
    width_bound: int = None,
    debug=False,
):
    td = bn.td
    if USING_COMPLEXITY_WIDTH:
        final_width_bound = weight_from_domain_size(width_bound)
    else:
        final_width_bound = width_bound
    selected, seen = find_subtree(td, budget, history, debug=False)
    history.update(selected)
    prep_tuple = prepare_subtree(bn, selected, seen, debug)
    if prep_tuple is None:
        return
    forced_arcs, forced_cliques, data, pset_acyc = prep_tuple
    # if debug:
    #     print("filtered data:-")
    #     pprint(data)
    old_score = bn.compute_score(seen)
    max_score = compute_max_score(data, bn)
    if RELAXED_PARENTS:
        # too strict
        # assert max_score + EPSILON >= old_score, "max score less than old score"
        assert round(max_score + EPSILON, 4) >= round(
            old_score, 4
        ), "max score less than old score"
        if max_score < old_score:
            print("#### max score smaller than old score modulo epsilon")
    cur_offset = sum(bn.offsets[node] for node in seen)
    if debug:
        print(
            f"potential max: {(max_score - cur_offset)/bn.best_norm_score:.5f}",
            end="",
        )
    if (max_score - cur_offset) / bn.best_norm_score <= LAZY_THRESHOLD:
        if debug:
            print(" skipping because lazy threshold not met")
        SOLUTION.skipped += 1
        return

    pos = dict()  # placeholder layout
    if not CLUSTER and debug:
        pos = graphviz_layout(bn.dag, prog="dot")
        nx.draw(bn.dag, pos, with_labels=True)
        plt.suptitle("entire dag")
        plt.show()
        nx.draw(bn.dag.subgraph(seen), pos, with_labels=True)
        plt.suptitle("subdag before improvement")
        plt.show()
    if debug:
        print("old parents:-")
        pprint({node: par for node, par in bn.parents.items() if node in seen})
    domain_sizes = DOMAIN_SIZES if USING_COMPLEXITY_WIDTH else None
    try:
        replbn = solve_bn(
            data,
            final_width_bound,
            bn.input_file,
            forced_arcs,
            forced_cliques,
            pset_acyc,
            timeout,
            domain_sizes,
            MAXSAT_SOLVER,
            debug,
        )
    except NoSolutionError as err:
        SOLUTION.nosolution += 1
        print(f"no solution found by maxsat, skipping\n(reason: {err})")
        return
    except UnsatisfiableInstanceError:
        SOLUTION.unsat += 1
        print(f"!!! maxsat instance unsatisfiable, skipping")
        return

    new_score = replbn.compute_score()
    if not CLUSTER and debug:
        nx.draw(replbn.dag, pos, with_labels=True)
        plt.suptitle("replacement subdag")
        plt.show()
    if debug:
        print("new parents:-")
        pprint(replbn.parents)
    if debug:
        print(f"score change: {old_score:.3f} -> {new_score:.3f}")
    if USE_COMPLEXITY_WIDTH:
        old_cw = compute_complexity_width(td, DOMAIN_SIZES, include=selected)
        new_cw = compute_complexity_width(replbn.td, DOMAIN_SIZES)
        old_acw = compute_complexity_width(
            td, DOMAIN_SIZES, include=selected, approx=True
        )
        new_acw = compute_complexity_width(replbn.td, DOMAIN_SIZES, approx=True)
        print(f"old: {old_cw}|{old_acw}\tnew: {new_cw}|{new_acw}")
    # replacement criterion
    if USING_COMPLEXITY_WIDTH and old_cw > width_bound:
        if new_cw > width_bound:
            return False
    elif USING_COMPLEXITY_WIDTH and new_score == old_score and new_cw > old_cw:
        return False
    elif (
        new_score < old_score
    ):  # in case not using cw, then this is the only check
        return False
    print(f"score change: {old_score:.3f} -> {new_score:.3f}, replacing ...")
    td.replace(selected, forced_cliques, replbn.td)
    # update bn with new bn
    bn.replace(replbn)
    if __debug__:
        bn.verify(verify_treewidth=not USING_COMPLEXITY_WIDTH)
    return True


@dataclass(repr=False, eq=False, unsafe_hash=True)
class SlimSolver:

    """
    allows to run slim in different modes
    assumes heuristic unchanged throughout
    """

    # fields
    filename: str
    start_treewidth: int
    solution: Solution  # safer to init externally
    seed: int = SEED
    heuristic: str = HEURISTIC
    debug: bool = False

    # don't need external init
    start_time: float = -1
    history: Counter = field(init=False, default=None)

    # for specific modes
    mimic_heur_proc: Popen = field(init=False, default=None)
    mimic_heur_running: bool = False
    mimic_heur_prev_score: float = None

    def elapsed(self):
        return now() - self.start_time

    def elapsed_msg(self):
        return f"(after {self.elapsed():.1f} seconds)"

    def reset_history(self, bn):
        self.history = Counter(dict.fromkeys(bn.td.decomp.nodes, 0))

    # different components of slim modularized into reusable chunks

    def pre_solve(self):
        """start timer and checkpointing"""
        self.start_time = now()
        if SAVE_AS and CHECKPOINT_MILESTONES:
            activate_checkpoints(lambda: self.solution.value, SAVE_AS)

    def read_starting_bn(self, start_with_file: str) -> TWBayesianNetwork:
        if not os.path.isfile(start_with_file):
            print(
                f"specified start-with file doesn't exist, quitting",
                file=sys.stderr,
            )
            return
        if self.debug:
            print(f"starting with {start_with_file}, not running heuristic")
        # todo[safety]: handle case when no heuristic solution so far
        # todo[safety]: make add_extra_tuples a cli option
        add_extra_tuples = self.heuristic in INTERLEAVABLE_HEURISTICS
        return parse_res(
            self.filename,
            self.start_treewidth,
            start_with_file,
            add_extra_tuples=add_extra_tuples,
            augfile=AUGMENTED_JKL,
        )

    def mimic_start_heuristic(self, start_with_bn=None):
        assert self.heuristic in ["hc", "hcp"], "selected heuristic not deterministic"

        start_sol_name = SAVE_AS.replace(".res", "_start4heur.res")
        heur_sol_name = SAVE_AS.replace(".res", "_heur.res")
        if start_with_bn is not None:
            # save bn to file for heuristic
            write_res(start_with_bn, start_sol_name, write_elim_order=True)
            print(f"[mimic] saving bn to file for heur to start with")
            start_with = start_sol_name
            self.mimic_heur_prev_score = start_with_bn.score
        else:
            start_with = None

        self.mimic_heur_proc = start_hcet_proc(
            DATFILE,
            self.start_treewidth,
            self.filename,
            self.heuristic,
            outfile=heur_sol_name,
            start_with=start_with,
            debug=False,
        )
        print(f"[mimic] started heur proc (pid: {self.mimic_heur_proc.pid})\n")
        self.mimic_heur_running = True

    def mimic_check_heuristic(self):
        slim_bn = self.solution.value
        heur_sol_name = SAVE_AS.replace(".res", "_heur.res")

        if self.mimic_heur_running:
            try:
                bn = check_hcet_proc(
                    self.mimic_heur_proc,
                    self.start_treewidth,
                    self.filename,
                    outfile=heur_sol_name,
                    debug=False,
                )
            except NoSolutionError as err:
                print(f"[mimic] heur proc resulted in error ({err})")
                print(f"[mimic] restarting heur with latest slim solution")
                self.mimic_start_heuristic(start_with_bn=slim_bn)
            else:
                if bn is not None:  # heuristic terminated normally
                    bn.verify()
                    # since new tuples could've been injected
                    # slim_bn.recompute_stats()
                    # slim_bn.recompute_score()
                    self.mimic_heur_running = False
                    # todo: decide if epsilon needed for following comparison
                    if bn.score > slim_bn.score:
                        print(
                            f"[mimic] heur score: {bn.score:.5f} "
                            f"(better) {self.elapsed_msg()}"
                        )
                        self.solution.update(bn)
                        self.mimic_heur_prev_score = bn.score
                        self.solution.jumps += 1
                    else:
                        print(
                            f"[mimic] heur score: {bn.score:.5f} "
                            f"(worse) {self.elapsed_msg()}"
                        )
                        # can potentially trigger the mimic_start_heuristic
                        # in the following if block

        # check if heuristic running
        if not self.mimic_heur_running:
            # start heuristic only if slim has made non-trivial improvement
            # if slim_bn.score > self.mimic_heur_prev_score + MIMIC_SCORE_THRESHOLD:
            # disabled this check due to hill-climbing's non-determinism
            self.mimic_start_heuristic(slim_bn)
            return  # either not running or just dispatched, so no need to check


    def run_heuristic(
        self, offset: int = None, max_time: int = None
    ) -> TWBayesianNetwork:
        """
        offset needed for blip heuristics
        max_time needed for mimic mode
        """
        if MIMIC and ANYTIME_HEURISTIC:
            assert offset is not None, "offset required for mimic mode"
            assert max_time is not None, "max_time required in mimic mode"
            if CONSTRAINED:
                raise NotImplementedError(
                    "mimic mode doesn't currently support constrained bnsl"
                )
            if self.debug:
                print("starting heuristic proc for mimicking")
            mimic_file = "temp-mimic.res"
            self.mimic_heur_proc = start_blip_proc(
                self.filename,
                self.start_treewidth,
                outfile=mimic_file,
                timeout=max_time,
                seed=self.seed,
                solver=self.heuristic,
                debug=False,
            )
            if self.debug:
                print(f"waiting {offset}s")
            sleep(offset)
            # todo[safety]: make more robust by wrapping in try except (race condition)
            return parse_res(self.filename, self.start_treewidth, mimic_file)
        else:
            if self.heuristic in INTERLEAVABLE_HEURISTICS:
                if self.debug:
                    print(f"running initial heuristic until termination")
                return run_hcet(
                    DATFILE, self.start_treewidth, self.filename, self.heuristic
                )
            else:
                assert offset is not None, "offset needed for BLIP heuristics"
                if self.debug:
                    print(f"running initial heuristic for {offset}s")
                return run_blip(
                    self.filename,
                    self.start_treewidth,
                    timeout=offset,
                    seed=self.seed,
                    solver=self.heuristic,
                )

    def run_heuristic_on(self, slim_sol_fname) -> TWBayesianNetwork:
        """
        run heuristic starting from given bn
        (only supported by INTERLEAVABLE_HEURISTICS)
        """

        # if MIMIC:  # no mimic possible
        assert (
            self.heuristic in INTERLEAVABLE_HEURISTICS
        ), f"only {INTERLEAVABLE_HEURISTICS} can start from supplied BN"

        if self.debug:
            print(f"running heuristic on given bn until termination")
        return run_hcet(
            DATFILE,
            self.start_treewidth,
            self.filename,
            self.heuristic,
            start_with=slim_sol_fname,
        )

    def post_heuristic(self, bn: TWBayesianNetwork):
        """verify and checkpoint if applicable"""
        if __debug__:
            bn.verify()
        # save checkpoint: milestone > start
        if CHECKPOINT_MILESTONES:
            write_res(
                bn, SAVE_AS.replace(".res", "_start.res"), write_elim_order=True
            )
        return bn

    def pre_slim(self, bn: TWBayesianNetwork, record_start_score=False):
        """record initial bn+stats, set randseed and reset history"""
        # record initial bn
        self.solution.update(bn)
        bn.recompute_score()
        prev_score = bn.score
        print(f"Starting score: {prev_score:.5f}")
        # if debug and DATFILE: print(f"Starting LL: {eval_ll(bn, DATFILE):.6f}")
        if record_start_score:
            self.solution.start_score = prev_score  # todo: think about this

        self.reset_history(bn)
        if self.seed:
            random.seed(self.seed)

    def post_slim(self):
        """stop mimic proc and print ending message"""
        if MIMIC and ANYTIME_HEURISTIC:
            if self.debug:
                print("stopping heur proc")
            stop_blip_proc(self.mimic_heur_proc)

        print(f"done {self.elapsed_msg()}")

    def pre_slimpass(self, bn: TWBayesianNetwork):
        """set width_bound (useful for extending to cwidth/constrained)"""
        width_bound = self.start_treewidth
        return bn.score, width_bound

    def post_slimpass(
        self,
        bn: TWBayesianNetwork,
        replaced: bool,
        prev_score: float,
        width_bound,
    ):
        """
        check for improvement, mimic borrow check and log relevant messages.
        """
        self.solution.num_passes += 1

        new_score = bn.score
        if new_score > prev_score:
            print(
                "*** New improvement!"
                f"{new_score:.5f} {self.elapsed_msg()} ***"
            )
            prev_score = new_score
            self.solution.update(bn)
            self.solution.num_improvements += 1
            if USE_COMPLEXITY_WIDTH and new_score >= self.solution.start_score:
                self.cw_stop_looping = True
        elif replaced:
            print("*** No improvement, but replacement performed ***")
            prev_score = new_score
            self.solution.update(bn)

        if MIMIC and ANYTIME_HEURISTIC:
            heur_score = check_blip_proc(self.mimic_heur_proc, debug=False)
            if heur_score > bn.score:
                if self.debug:
                    print(
                        f"heuristic solution better {heur_score:.5f} > {bn.score:.5f}, mimicking"
                    )
                self.solution.restarts += 1
                newbn = parse_res(self.filename, self.start_treewidth, outfile)
                new_score = newbn.score
                assert abs(
                    new_score >= heur_score - 1e-5
                ), f"score exaggerated, reported: {heur_score}\tactual score: {new_score}"
                bn = newbn
                prev_score = new_score
                self.solution.update(bn)
                # reset history because fresh tree decomposition
                self.reset_history(bn)

        if self.debug:
            print(
                f"* Iteration {self.solution.num_passes}:\t"
                f"{bn.score:.5f} {self.elapsed_msg()}\n"
            )


    def just_slim(
        self,
        budget: int = BUDGET,
        sat_timeout: int = TIMEOUT,
        max_passes: int = MAX_PASSES,
        max_time: int = MAX_TIME,
        start_with_bn: TWBayesianNetwork = None,
        start_with_file: str = None,
    ):
        """run slim on provided starting bn until total duration of max_time"""

        # read in provided starting bn
        if start_with_bn is not None:
            start_bn = start_with_bn
        elif start_with_file is not None:
            start_bn = self.read_starting_bn(start_with_file)
        else:
            raise ValueError(
                "starting bn must be provided either directly or as file"
            )

        # bn = self.post_heuristic(bn)  # todo: check if needed
        self.pre_slim(start_bn, record_start_score=True)

        # start heuristic proc for mimic mode, if needed
        if MIMIC and not ANYTIME_HEURISTIC:
            self.mimic_start_heuristic(start_with_bn=start_bn)

        # repeated code begin
        while max_passes < 0 or self.solution.num_passes <= max_passes:
            bn = self.solution.value
            prev_score, width_bound = self.pre_slimpass(bn)

            # main forwarding call to slimpass
            replaced = slimpass(
                bn,
                budget,
                sat_timeout,
                self.history,
                width_bound,
                debug=False,
            )

            if replaced is None:  # no change by slimpass
                continue  # don't count this as a pass

            self.post_slimpass(bn, replaced, prev_score, width_bound)

            if MIMIC and not ANYTIME_HEURISTIC:  # sync with heuristic
                self.mimic_check_heuristic()

            if self.elapsed() > max_time:
                if self.debug:
                    print("time limit exceeded, quitting")
                break
        else:
            if self.debug:
                print(f"{max_passes} passes completed, quitting")
        # repeated code end

        # self.post_slim()  # todo: most likely don't need, confirm

    def heur_then_slim(
        self,
        budget: int = BUDGET,
        sat_timeout: int = TIMEOUT,
        max_passes: int = MAX_PASSES,
        max_time: int = MAX_TIME,
        offset: int = OFFSET,
    ):
        """
        run heuristic for offset amount of time and then slim until total
        duration of max_time
        """

        # run initial heuristic
        bn = self.run_heuristic(
            offset,
            max_time if (MIMIC and ANYTIME_HEURISTIC) else None
        )

        # checkpointing
        bn = self.post_heuristic(bn)

        self.pre_slim(bn)

        # check for one-shot slim
        if budget >= bn.dag.number_of_nodes():  # entire instance within budget
            print(
                "budget can handle entire instance, dispatching single SAT call"
            )
            sat_timeout = max_time
            max_passes = 1

        # repeated code begin
        while max_passes < 0 or self.solution.num_passes <= max_passes:
            prev_score, width_bound = self.pre_slimpass(bn)

            # main forwarding call to slimpass
            replaced = slimpass(
                bn,
                budget,
                sat_timeout,
                self.history,
                width_bound,
                debug=False,
            )

            if replaced is None:  # no change by slimpass
                continue  # don't count this as a pass

            self.post_slimpass(bn, replaced, prev_score, width_bound)

            if self.elapsed() > max_time:
                if self.debug:
                    print("time limit exceeded, quitting")
                break
        else:
            if self.debug:
                print(f"{max_passes} passes completed, quitting")
        # repeated code end

        self.post_slim()

    def _until_improvement(
        self,
        budget: int,
        sat_timeout: int,
        bn: TWBayesianNetwork,
        max_time: int,
    ):
        """run slim until it sees one improvement, or it exhausts max_time"""
        initial_improvements = self.solution.num_improvements

        personal_start = self.elapsed()
        while True:
            personal_elapsed = self.elapsed() - personal_start
            print(f"[interleaved] {personal_elapsed:.2f}s / {max_time}s")
            if personal_elapsed > max_time:
                if self.debug:
                    print("max_time limit exceeded while awaiting improvement")
                return False

            prev_score, width_bound = self.pre_slimpass(bn)

            # main forwarding call to slimpass
            replaced = slimpass(
                bn,
                budget,
                sat_timeout,
                self.history,
                width_bound,
                debug=False,
            )

            if replaced is None:  # no change by slimpass
                continue  # don't count this as a pass

            self.post_slimpass(bn, replaced, prev_score, width_bound)
            if self.solution.num_improvements > initial_improvements:
                return True

    def interleaved_slim(
        self,
        budget: int = BUDGET,
        sat_timeout: int = TIMEOUT,
        max_time: int = MAX_TIME,
        interleave_timeout: int = INTERLEAVE_TIMEOUT,
        start_with: str = None,
    ):
        """max_time acts as global timeout"""
        slim_sol_name = SAVE_AS.replace(".res", "_slim2heur.res")

        # flag to check if its the first heuristic run (no starting bn then)
        fresh_start = True

        while True:  #  runs until max_time reached
            if fresh_start:  # check if first loop
                # if initial heuristic solution externally provided
                if start_with is not None:
                    bn = parse_res(
                        self.filename,
                        self.start_treewidth,
                        start_with,
                        add_extra_tuples=True,
                    )
                else:
                    bn = self.run_heuristic()
                fresh_start = False

                # record initial heuristic score
                self.solution.start_score = bn.score
            else:
                bn = self.run_heuristic_on(slim_sol_name)

            bn = self.post_heuristic(bn)
            print(
                f"[interleaved] heur score: {bn.score:.2f} {self.elapsed_msg()}"
            )

            interleave_start = self.elapsed()
            remaining = interleave_timeout

            # keep running slim until interleave timeout reached
            while True:
                self.pre_slim(bn)
                # todo: skipping one-shot check here, is it ok?

                # at least one improvement
                improvement_found = self._until_improvement(
                    budget, sat_timeout, bn, remaining
                )
                if not improvement_found:
                    print("no new slim improvements, still interleaving")
                    # return
                    break

                # check if enough time remaining for another slim pass
                interleave_duration = self.elapsed() - interleave_start
                remaining = interleave_timeout - interleave_duration
                # account for next slimpass duration (better to end early?)
                if remaining < sat_timeout:
                    print("interleave_timeout will be exceeded, interleaving")
                    break

            print(
                f"[interleaved] slim score: {bn.score:.2f} {self.elapsed_msg()}"
            )

            if self.elapsed() > max_time:
                if self.debug:
                    print("max_time limit exceeded, quitting")
                break

            # save slim solution to file for heuristic
            write_res(self.solution.value, slim_sol_name, write_elim_order=True)
            print(
                "saving interleave checkpoint to",
                os.path.basename(slim_sol_name),
                "(for passing to heuristic)",
                end="\n\n",
            )


class SolverInterrupt(BaseException):
    pass


def term_handler(signum, frame):
    print(f"#### received signal {signum}, stopping...")
    raise SolverInterrupt


def register_handler():
    signums = [
        signal.SIGHUP,   # 1
        signal.SIGINT,   # 2
        signal.SIGTERM,  # 15
        signal.SIGUSR1,  # 10
        signal.SIGUSR2,  # 12
    ]
    for signum in signums:
        signal.signal(signum, term_handler)


def wandb_configure(wandb: wandb, args):
    basename, ext = os.path.splitext(os.path.basename(args.file))
    wandb.config.instance = basename
    wandb.config.treewidth = args.treewidth
    wandb.config.budget = args.budget
    wandb.config.sat_timeout = args.sat_timeout
    wandb.config.heuristic = args.heuristic
    wandb.config.offset = args.offset
    wandb.config.threshold = args.lazy_threshold
    wandb.config.seed = args.random_seed
    wandb.config.method = (
        args.heuristic if args.compare else f"slim_{args.heuristic}"
    )
    wandb.config.traversal = args.traversal_strategy
    wandb.config.relaxed = True  # int(args.relaxed_parents)
    wandb.config.mimic = int(args.mimic)
    wandb.config.datfile = args.datfile
    wandb.config.complexity_width = args.complexity_width
    wandb.config.cw_strategy = args.cw_strategy
    # process config
    wandb.config.platform = "cluster" if CLUSTER else "workstation"
    wandb.config.jobid = int(os.environ.get("MY_JOB_ID", -1))
    wandb.config.taskid = int(os.environ.get("MY_TASK_ID", -1))


# ArgumentParser setup
# noinspection PyTypeChecker
parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter
)
parser.add_argument("file", help="path to input file")
parser.add_argument("treewidth", help="bound for treewidth", type=int)

slim_args = parser.add_argument_group("SLIM Parameters")
slim_args.add_argument(
    "-b",
    "--budget",
    type=int,
    default=BUDGET,
    help="budget for size of local instance",
)
slim_args.add_argument(
    "-s",
    "--sat-timeout",
    type=int,
    default=TIMEOUT,
    help="timeout per MaxSAT call in seconds",
)
slim_args.add_argument(
    "-p",
    "--max-passes",
    type=int,
    default=MAX_PASSES,
    help="max number of passes of SLIM to run",
)
slim_args.add_argument(
    "-t",
    "--max-time",
    type=int,
    default=MAX_TIME,
    help="max time in seconds for SLIM to run",
)
slim_args.add_argument(
    '--maxsat-solver',
    type=str,
    choices=["nuwls-c", "uwrmaxsat10", "uwrmaxsat15"],
    default=MAXSAT_SOLVER,
    help="maxsat solver to use",
)
parser.add_argument(
    "-u",
    "--heuristic",
    default=HEURISTIC,
    choices=["kg", "ka", "kmax", "hc", "hcp"],
    help="heuristic solver to use",
)
parser.add_argument(
    "-o",
    "--offset",
    type=int,
    default=OFFSET,
    help="duration for which initial heuristic is run, "
    "doesn't apply to hc or hcp",
)
mode_args = parser.add_argument_group("Modes", "different modes of bn-slim")
mode_args.add_argument(
    "-c",
    "--compare",
    action="store_true",
    help="run only heuristic to gather stats for comparison",
)
mode_args.add_argument(
    "-m",
    "--mimic",
    action="store_true",
    help="mimic heuristic if it outperforms",
)
io_args = parser.add_argument_group(
    "I/O Arguments", "control input/output behaviour"
)
io_args.add_argument(
    "--start-with",
    default=None,
    help="optionally skip running heuristic and start with this solution",
)
io_args.add_argument(
    "--save-as", default="", help="filename to save final network as"
)
io_args.add_argument(
    "--checkpoint-milestones",
    action="store_true",
    help="save key milestone networks",
)
io_args.add_argument(
    "-v", "--verbose", action="store_true", help="verbose mode"
)
parser.add_argument(
    "-d",
    "--datfile",
    default=None,
    help="path to datfile, if omitted, complexity-width will not be tracked "
    "[required when --heuristic is hc or hcp and --start-with is not provided]",
)
cwidth_args = parser.add_argument_group(
    "for Bounded Complexity Width BNSL",
    "also referred to as bounded state space BNSL",
)
cwidth_args.add_argument(
    "-w",
    "--complexity-width",
    action="store_true",
    help="minimizing complexity width becomes main objective\n"
    "[requires option -d|--datfile]",
)
cwidth_args.add_argument(
    "--cw-strategy",
    default=CW_TRAV_STRAT,
    choices=["max", "max-rand", "max-min", "tw-max-rand"],
    help="complexity width reduction traversal strategy\n"
    "[ignored if option -w|--complexity width not provided]",
)
mode_args.add_argument(
    "-i",
    "--interleaved",
    action="store_true",
    help="run slim in interleaved mode, "
    "i.e., alternate between slim and heuristic "
    f"(only allowed when --heuristic in {INTERLEAVABLE_HEURISTICS}"
)
parser.add_argument(
    "--interleave-timeout",
    type=int,
    default=INTERLEAVE_TIMEOUT,
    help="minimum time (in s) to run each interleaved phase of slim for",
)
parser.add_argument(
    "-r",
    "--random-seed",
    type=int,
    default=SEED,
    help="random seed (set 0 for no seed)",
)

internal_args = parser.add_argument_group(
    "Internals", "parameters for controlling low-level inner workings"
)
internal_args.add_argument(
    "-z",
    "--lazy-threshold",
    type=float,
    default=LAZY_THRESHOLD,
    help="threshold below which to not try to improve local instances (set 0 to disable)",
)
internal_args.add_argument(
    "-x",
    "--relaxed-parents",
    action="store_true",
    help="(deprecated) relax allowed parent sets for maxsat encoding\n"
    "[warning: use at own risk, could terminate abruptly]",
)
internal_args.add_argument(
    "-y",
    "--traversal-strategy",
    default=TRAV_STRAT,
    choices=["random", "post", "pre"],
    help="td traversal strategy",
)

logging_args = parser.add_argument_group("Logging arguments")
logging_args.add_argument(
    "-l", "--logging", action="store_true", help="wandb logging"
)
logging_args.add_argument(
    "--project-name", default="twbnslim-test", help="wandb project name"
)

if __name__ == "__main__":
    args = parser.parse_args()
    filepath = os.path.abspath(args.file)
    LAZY_THRESHOLD = args.lazy_threshold
    RELAXED_PARENTS = True  # args.relaxed_parents
    MIMIC = args.mimic

    HEURISTIC = args.heuristic
    if HEURISTIC in ["kg", "ka", "kmax"]:
        ANYTIME_HEURISTIC = True
    elif HEURISTIC in ["hc", "hcp"]:
        ANYTIME_HEURISTIC = False

    SAVE_AS = os.path.abspath(args.save_as) if args.save_as else ""
    DOMAIN_SIZES = get_domain_sizes(args.datfile) if args.datfile else None
    if args.complexity_width:
        USE_COMPLEXITY_WIDTH = True
        if DOMAIN_SIZES is None:
            parser.error("--complexity-width requires --datfile")
    CW_TRAV_STRAT = args.cw_strategy
    MAXSAT_SOLVER = args.maxsat_solver
    CHECKPOINT_MILESTONES = args.checkpoint_milestones
    if CHECKPOINT_MILESTONES and not SAVE_AS:
        parser.error("--checkpoint-milestones switch requires --save-as option")
    INTERLEAVED = args.interleaved
    if INTERLEAVED and (
        args.heuristic not in INTERLEAVABLE_HEURISTICS
        or USE_COMPLEXITY_WIDTH
    ):
        parser.error(
            "--interleaved only supported when --heuristic in"
            f"{INTERLEAVABLE_HEURISTICS}, "
            "and not --complexity-width mode"
        )
    if INTERLEAVED and not SAVE_AS:
        parser.error("-i/--interleaved mode requires --save-as option")

    if args.budget <= args.treewidth and not args.compare:
        print("budget smaller than treewidth bound, quitting", file=sys.stderr)
        sys.exit()
    print("not" * __debug__, "running optimized")
    if args.start_with is not None:
        START_WITH = os.path.abspath(args.start_with)
    if (
        args.datfile is None
        and args.heuristic in ["hc", "hcp"]
        and START_WITH is None
    ):
        parser.error(
            "--datfile required when heuristic is hc/hcp and"
            "--start-with not provided"
        )
    if MIMIC and START_WITH is None:
        parser.error("--start-with required for --mimic")
    if MIMIC and not SAVE_AS:
        parser.error("--mimic mode requires --save-as option")
    DATFILE = args.datfile

    TRAV_STRAT = args.traversal_strategy
    logger = lambda x: x  # no op
    # logger = lambda x: print(f"log: {x}")  # local log
    LOGGING = args.logging
    SEED = args.random_seed
    if LOGGING:
        print("initializing wandb...")
        wandb.init(project=args.project_name)
        print("initialized wandb")
        wandb_configure(wandb, args)
        logger = wandb.log
    SOLUTION.logger = logger

    # swap out jkl file with a copy to allow safe augmentation
    shutil.copyfile(filepath, AUGMENTED_JKL)
    original_file = filepath
    filepath = AUGMENTED_JKL

    # if comparison requested, compare then exit
    if args.compare:
        outfile = SAVE_AS or "temp.res"
        logger = lambda x: print(f"log: {x}")  # local log
        monitor_blip(
            filepath,
            args.treewidth,
            logger,
            timeout=args.max_time,
            seed=SEED,
            outfile=outfile,
            solver=args.heuristic,
            domain_sizes=DOMAIN_SIZES,
            save_as=SAVE_AS,
            debug=args.verbose,
        )
        sys.exit()

    # initialize slim solver (used for all the passes)
    solver = SlimSolver(
        filepath,
        args.treewidth,
        SOLUTION,
        SEED,
        args.heuristic,
        args.verbose,
    )

    register_handler()

    try:
        # start timer and checkpointing
        solver.pre_solve()

        # if interleaved mode (can be sure not in cwidth mode)
        if INTERLEAVED:
            solver.interleaved_slim(
                budget=args.budget,
                sat_timeout=args.sat_timeout,
                max_time=args.max_time,
                interleave_timeout=args.interleave_timeout,
                start_with=START_WITH,
            )
        else:
            # run slim once (optionally run heuristic before)
            # if no starting network provided, run heuristic first
            if START_WITH is None:
                res = solver.heur_then_slim(
                    args.budget,
                    args.sat_timeout,
                    args.max_passes,
                    args.max_time,
                    args.offset,
                )
            else:  # otherwise, run slim directly
                # (if applicable, mimic runs here)
                res = solver.just_slim(
                    args.budget,
                    args.sat_timeout,
                    args.max_passes,
                    args.max_time,
                    start_with_file=START_WITH,
                )
    except SolverInterrupt:
        print("solver interrupted")
    finally:
        if SOLUTION.value is None:
            print("terminated. no solution computed so far!")
        else:
            # verify final bn (not required if optimizing complexity width)
            # todo[req]: complexity width separate verification
            SOLUTION.value.verify(verify_treewidth=not USE_COMPLEXITY_WIDTH)
            print("verified")
            if SAVE_AS:
                save_fname = SAVE_AS.replace(".res", "_final.res")
                write_res(SOLUTION.value, save_fname, write_elim_order=True)
                print(
                    "saving final network to", save_fname, "as final checkpoint"
                )
            valid_passes = SOLUTION.num_passes - SOLUTION.skipped
            success_rate = (
                SOLUTION.num_improvements / valid_passes if valid_passes else 0
            )
            treewidths = dict(
                start_tw=args.treewidth,
                final_tw=SOLUTION.value.td.compute_width(),
            )
            if LOGGING:
                wandb.log({"success_rate": success_rate})
                wandb.log(treewidths)
                if SOLUTION.num_improvements > 0:
                    wandb.log({"improved": True})
            else:
                print("final metrics:")
                pprint(SOLUTION.data)
                print(f"start score: {SOLUTION.start_score:.5f}")
                print(f"success_rate: {success_rate:.2%}")
                print(f"final score: {SOLUTION.value.score:.5f}")
                print(treewidths)
                if DOMAIN_SIZES:
                    print(
                        "complexity-width:",
                        compute_complexity_width(
                            SOLUTION.value.td, DOMAIN_SIZES
                        ),
                    )
